/***************************************************************************
                          astroaspects.h  -  description
                             -------------------
    begin                : Mon Feb 7 2001
    copyright            : (C) 2001 by Christophe Gros
    email                : skylendar@yahoo.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef ASTROASPECTS_H
#define ASTROASPECTS_H

#include "../config.h"

#include "astrodefs.h"
#include "astrolist.h"
#include "astrostring.h"
#include "astroresources.h"

#define AL(a, b) Already[(Max_x + 1) * b + a]

class AstroObjs;
class AstroRestrictions;
class _AstroRestrictions;
class Values;

struct AspectsConstants
{
  double Angle;
  enum AspectKind Kind;
  int Harmonic;
};

struct AspectRestrictions
{
  bool Rest;
  double Orb;
  int Color, Influence;
};

extern const struct AspectsConstants*  AspConstVector;
#define I enum Aspects i

/* The accuracy of an aspect
     Exact - Good: less than or equal 30% of the orb
     Medium      : less than of equal 60% of the orb
     Poor        : more than 60% of the orb
  */
enum AspectAccuracy { Exact, Good, Medium, Poor, Applying, Separating };

typedef unsigned char asp;

struct MidPoint
{
	double Mp; // Md-point value
	enum Objs Oa, Ob; // Corresponding objects
};

class _AstroAspects
{
	public:

	_AstroAspects(_AstroRestrictions* ar);
	inline double operator * (I)   const { return AspConstVector[i].Angle; }
  inline enum AspectKind Kind(I) const { return AspConstVector[i].Kind; }
  inline int Harmonic(I)         const { return AspConstVector[i].Harmonic; }
  inline bool operator == (I)    const { return !Asprst[i].Rest; }
  inline double operator / (I)   const { return Asprst[i].Orb; }
	inline int Color(I)            const { return Asprst[i].Color; }
  inline int Influence(I)        const { return Asprst[i].Influence; }
  inline const struct Rest* GetRest() const { return Restr; }

  enum Aspects IsAspect(enum Objs oa, const Values& a,
  	enum Objs ob, const Values& b, enum AspectAccuracy& side) const ;

  protected:

  struct Rest* Restr;
  struct AspectRestrictions* Asprst;
  bool IsMinor;
};

/** This class manages the aspects data */
class AstroAspects : public _AstroAspects
{
 public:
  
  AstroAspects(AstroRestrictions* ar);

  inline const AstroString& operator << (I) const
    { return *Asr->AspectNames[i]; }
  inline const AstroString& operator  < (I) const
    { return *Asr->AspectShortNames[i]; }
  char GetAspChar(I) const;

  /** Returns the Aspect if objs oa and ob are in the corresponding aspect.
  @param oa Object a
  @param ob Object b
  @param a AstroObjs a
  @param n AstroObjs b
  */
  enum Aspects IsAspect(enum Objs oa, const AstroObjs& a,
  	enum Objs ob, const AstroObjs& b, enum AspectAccuracy& side) const;

 private:

  bool IsMinor;
  char *AspChars;
};

/** class provided by  AstroAspectsList
    It contains the values of a found aspect
    Angle_a, Angle_b: angles (in degrees) of the objs Oa and Ob.
    Strength: the power of the aspect (Positive number) */

class AspectFound
{
 public:

  /** Constructor */
  AspectFound(const _AstroAspects& aa, enum Aspects asp, enum Objs oa, const Values& a,
  	enum Objs ob, const Values& b, enum AspectAccuracy side);

  enum Aspects Asp; // Found aspect.
  enum Objs Oa, Ob; // Involved objects.
  double Angle_a, Angle_b, DiffLong, Orb; // Angles of involved objects + Difference of longitude + actual orb.
  /** Returns the accuracy of the aspect
      @return AspectAccuracy
  */
  enum AspectAccuracy GetAccuracy();

  // If an aspect is exact, applying or separating
  enum AspectAccuracy AspectSide;

 private:

  _AstroAspects* Aa; // Aspects table pointer.
};

/** This class is the base class for the aspects computing. It only maintains
	a table of found aspects */
class AstroFoundAspectsTable
{
	public:

	/** constructor */
	AstroFoundAspectsTable() { Already = 0; Max_x = Max_y = Sun; }

	/** destructor */
	virtual ~AstroFoundAspectsTable() { if(Already) delete Already; Max_x = Max_y = Sun; }

	/** Gets a table of found aspects
	@param aa an aspects table reference
	@param a, b the regarded objects values
	@param ifsameplanet if the aspects to same planets are regarded
	@param ar, br the restrictions tables
*/
	void GetTable(const _AstroAspects& aa, const Values& a, const Values& b, bool ifsameplanet,
		const _AstroRestrictions& ar, const _AstroRestrictions& br);

	/* Gets the aspect between x and y directly from the Aspects table */
	inline enum Aspects Get(int x, int y) const { return (enum Aspects) AL(x, y); }

	/** Returns the maximum ojects indexes */
	inline void GetBounds(int* max_x, int* max_y) const { *max_x = Max_x; *max_y = Max_y; }

protected:

	void TestAspects(const _AstroAspects& aa, const Values& a, const Values& b,
		const _AstroRestrictions& ar, const _AstroRestrictions& br);

	void Reset(int a, int b);

	virtual void PutAspect(const _AstroAspects& aa, enum Aspects asp, enum Objs oa,
		const Values& a, enum Objs ob, const Values& b, enum AspectAccuracy side) {}

	bool IfSamePlanet;
	asp* Already;
	enum Objs Max_x, Max_y;
};

/** This class holds a list of found aspects between AstroObjs a and b */
class AstroFoundAspectsList: public AstroFoundAspectsTable, public AstroList(AspectFound)
{
 public:

   /** Constructor */
  AstroFoundAspectsList() { NbMP = 0; MidP = 0; }

	/** Destructor */
	~AstroFoundAspectsList() { if(MidP) delete MidP; }

  /** Constructor.
   @param a ref to first objects set
   @param b ref to second objects set
   @param ifsameplanet If TRUE regard the possible aspects between the same planet, (e.g. Venus in both a and b)
	 @param ifmidp also compute mid-points */

  inline AstroFoundAspectsList(const AstroAspects& aa, const AstroObjs& a, const AstroObjs& b, bool ifsameplanet)
  { GetList(aa, a, b, ifsameplanet); }

  /** Same as above, but not constructor method. */
  void GetList(const AstroAspects& aa, const AstroObjs& a, const AstroObjs& b, bool ifsameplanet);

	/** Computes the mid-points list
	@param a
	@param b Corresponding astroobjs
	@param ifsameplanet if mid-point for the same planet
	*/
  void ComputeMidP(const AstroObjs& a, const AstroObjs& b, bool ifsameplanet);

  /** Returns the Mid-Point value of the ith item */
	inline const struct MidPoint& operator * (int i) const { return MidP[i]; }

	int NbMP;

 private:

  void TestAspects(const AstroAspects& aa, const AstroObjs& a, const AstroObjs& b);
	void GetMidP(double d, const AstroObjs& b, int i, int* k);
	void PutAspect(const _AstroAspects& aa, enum Aspects asp, enum Objs oa,
		const Values& a, enum Objs ob, const Values& b, enum AspectAccuracy side);

	struct MidPoint* MidP;
};

#endif
  
