/***************************************************************************
                          astrochart.cpp  -  description
                             -------------------
    begin                : Mon Oct 16 2000
    copyright            : (C) 2000 by Christophe Gros
    email                : skylendar@yahoo.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "astrochart.h"
#include "astroparams.h"
#include "rawtextchart.h"
#include "stdwheelchart.h"
#include "stdgraphiclist.h"
#include "aspscanlist.h"
#include "astroprocess.h"
#include "astrosql.h"
#include "astroaspects.h"

AstroChart::AstroChart(bool owninterface)
{
  int i;
  for(i = 0; i < NBDATA; i++)
    Ov[i] = 0;
  Acb = 0;
  Cbp = new ChartBaseParams(SC(ChartParams *,Asr), owninterface ? CreateInterface() : Asf);
  CHECK_PTR(Cbp);
  Ahl->Append(this);
}

AstroChart::~AstroChart()
{
  int i;
  if(Acb) delete Acb;
  for(i = 0; i < NBDATA; i++)
    if(Ov[i])
      delete Ov[i];
  Cbp->Aif->Kill();
  Ahl->Remove(this);
}

/** This method displays the chart according to
the Chart Type */
bool AstroChart::Display()
{
  if(Acb)
  {
    if(ct != Acb->ct)
    {
      delete Acb;
      CreateChart();
    }
    else
      return Acb->Redisplay();
  }
  else
    CreateChart();
  return TRUE;
}

/** This method creates an Acb according to ct */
void AstroChart::CreateChart()
{
  switch(ct)
  {
    case RawText     : Acb = new RawTextChart (Cbp, &Ov); break;
    case StdWheels   : Acb = new StdWheelChart(Cbp, &Ov); break;
    case GraphicList : Acb = new StdGraphicList(Cbp, &Ov); break;
		case AspScList   : Acb = new AspScanList(Cbp, &Ov); break;
  };
  Acb->PostRecompute();
  Acb->Redisplay();
}

void AstroChart::Repaint(bool b)
{
	if(Acb && ct == AspScList)
		Acb->PostRecompute();
	if(b)
		Acb->IfRedrawChart = TRUE;
}

void AstroChart::SetData(AstroData *ad, int i, bool placeonly)
{
  if(i < 0 || i >= NBDATA)
    return;
  if(!Ov[i])
  {
    Ov[i] = new AstroObjs(*ad, this, i);
    CHECK_PTR(Ov[i]);
    Ov[i]->SetInterface(Cbp->Aif);
    if(Acb)
      Acb->PostSetData(i);
  }
  else
  {
    AstroData* adp = SC(AstroData *,Ov[i]);
    if(!placeonly)
      *adp = *ad;
    adp->SetInterface(Cbp->Aif);
    adp->Recompute();
  }
}

int AstroChart::IfData(const AstroData* ad, bool placeonly) const
{
  int i;
  for(i = 0; i < NBDATA; i++)
    if(Ov[i] && (placeonly ? (Ov[i]->PlaceIdx == ad->PlaceIdx) : (Ov[i]->Idx == ad->Idx)))
      return i;
  return -1;
}

bool AstroChart::EditChartParams()
{
  int i;
  if(Cbp->Edit())
  {
    for(i = 0; i < NBDATA; i++)
      if(Ov[i])
      {
        *SC(ChartParams*, Ov[i]) = *SC(ChartParams*, Cbp);
        Ov[i]->Recompute();
      }
  }
  return TRUE;
}

int AstroChart::IfRestrict(const AstroRestrictions* ar) const
{
    int i;
    for(i = 0; i < NBDATA; i++)
     if(Ov[i] && ~*Ov[i] && (~*Ov[i])->Idx == ar->Idx)
       return i;
   return -1;
}

bool AstroChart::UpdateRestrict(AstroRestrictions *ar)
{
  int i;
  bool ret = FALSE;
  for(i = 0; i < NBDATA; i++)
    if(Ov[i] && ~*Ov[i] && (~*Ov[i])->Idx == ar->Idx)
    {
      SetRestrict(ar, i);
      ret = TRUE;
    }
  return ret;
}

bool AstroChart::UpdateData(AstroData *ad, bool placeonly)
{
  int i = IfData(ad, placeonly);
  if(i != -1)
  {
    SetData(ad, i);
    return TRUE;
  }
  return FALSE;
}

bool AstroChart::UpdateZif(Index idx, const char* zif)
{
  int i;
  bool ret = FALSE;
  for(i = 0; i < NBDATA; i++)
    if(Ov[i] && Ov[i]->CountryIdx == idx)
    {
      Ov[i]->SetTZ(zif);
      Ov[i]->Recompute();
      ret = TRUE;
    }
  return ret;
}

void AstroChart::SetRestrict(AstroRestrictions* ar, int i)
{
  if(i < 0 || i >= NBDATA) return;
  if(Ov[i])
  {
    Ov[i]->SetRestrict(ar, i);
    Cbp->Aif->SetRestrict(ar, i);
    Ov[i]->Recompute();
    Acb->IfRedrawChart = TRUE;
  }
  else
    AstroMessage(i18n("Can't set this restrictions set for an empty ring"));
}

int AstroChart::NbSingle()
{
  int i, s = 0;
  for(i = 0; i < NBDATA; i++)
    if(Ov[i] && Ov[i]->GetSubType() == Single)
      s++;
  return s;
}

bool AstroChart::Show(int i)
{
  int j, nb = 0;
  AstroObjs& ao = *Ov[i];
  if(ao.Show)
  {
    for(j = 0; j < NBDATA; j++)
      if(Ov[j] && Ov[j]->Show)
	nb++;
    if(nb == 1)
      return FALSE;
  }
  ao.Show = !ao.Show;
  Acb->IfRedrawChart = TRUE;
  return TRUE;
}

bool AstroChart::SetSubType(ChartSubType cst, int i)
{
  int nbs = NbSingle();
  if(i < 0 || i >= NBDATA) return FALSE;
  if(!Ov[i])
    SetData(Adl->Get(0), i, TRUE);
  AstroObjs& ao = *Ov[i];
  switch(cst)
  {
    case NoSt:      break;
    case Show_Hide: return Show(i); break;
    case Single:    ao.SetSubChart(cst); ao.Recompute(); break;
    case Transit:
    case Direction:
    case Progression:
      if(nbs == 0 || (nbs == 1 && cst != Single && cst != Show_Hide && ao.GetSubType() == Single))
     	  return FALSE;
      else
      {
       	ao.SetSubChart(cst);
       	ao.Recompute();
      }
      break;
  }
  return TRUE;
}

Index AstroChart::ExtractandSave(int i, Index& oldidx, bool& createnew, const AstroFoundAspectsList* afal)
{
  AstroRequest arq;
  AstroString as;
  int j;
  char c;
  if(i < 0 || i >= NBDATA || !Ov[i]) return 0;
  AstroObjs &ao = *Ov[i], &fs = *(*this)[Acb->Search()];
  Index restidx, f = fs.Idx;
  oldidx = ao.Idx;
  restidx = (~ao)->Idx;
  if(restidx >= BOUND)
     restidx = 0;
  createnew = FALSE;
  if(ao.Idx == 0) return 0;
  if(fs.Modified || fs.Idx >= BOUND)
  {
    AstroMessage(Sprintf(i18n("The Data set %a should be saved\n before extracting these values"), &as, &fs.Name));
    return 0;
  }
  switch(ao.GetSubType())
    {
      case Single :     c = 'S'; break;
      case Transit:     c = 'T'; break;
      case Direction:   c = 'D'; break;
      case Progression: c = 'P'; break;
      default:          c = 'X';
    };
  if(ao.Modified || ao.Idx >= BOUND)
  {
    if(c != 'S' && ao.Name.isEmpty())
    {
      int k = 0;
      AstroData* ad = new AstroData(ao);
      ad->Idx = 1;
      arq.Exec("SELECT MAX(Idx) FROM Data");
      if(arq.Valid())
        k = arq.GetInt("Max") + 1;
      ad->Name = Sprintf("%a #%d on %a", &as, &ao.GetSubTypeName(), k, &fs.Name);
      Adl->AddData(ad);
      createnew = TRUE;
      ad->DbSave();
      SetData(ad, i);
    }
    else
      ao.DbSave();
  }
  arq.Exec("SELECT DataIdx,Type FROM VDV WHERE Idx = %d", ao.Idx);
	arq.Begin();
  if(arq.Valid() && arq.GetChar("Type") == c)
  {
     arq.Exec("UPDATE VData SET RefIdx = %d, RestIdx = %d, Type = '%c' WHERE DataIdx = %d",
           f,restidx, c, ao.Idx);
     arq.Exec("DELETE FROM Obj WHERE DataIdx = %d", ao.Idx);
     arq.Exec("DELETE FROM Asp WHERE DataIdx = %d", ao.Idx);
  }
  else
  {
    arq.Exec("INSERT INTO VData (DataIdx, RefIdx, RestIdx, Type) VALUES (%d, %d, %d, '%c')",
         ao.Idx, f, restidx, c);
  }
   for(j = Sun; j <= Last_Planet; j++)
     if(ao == j)
       arq.Exec("INSERT INTO Obj (DataIdx, Id, Lon, Lat, Retro, InHouse) VALUES (%d, '%a', %f, %f, %b, %d)",
         ao.Idx, &(ao < j), ao * j, (*ao[j])[1], ao[j]->Retro, ao.InHouse(j));
   if(ao.IfHouses())
   {
     for(j = Ascendant; j <= House_12; j++)
       if(ao == j)
         arq.Exec("INSERT INTO Obj (DataIdx, Id, Lon, Lat, Retro, InHouse) VALUES (%d, '%a', %f, %f, %b, %d)",
           ao.Idx, &(ao < j), ao * j, (*ao[j])[1], ao[j]->Retro, ao.InHouse(j));

   }
   if(ao.IfExotic())
   {
     for(j = First_Exotic; j <= Last_Exotic; j++)
       if(ao == j)
         arq.Exec("INSERT INTO Obj (DataIdx, Id, Lon, Lat, Retro, InHouse) VALUES (%d, '%a', %f, %f, %b, %d)",
           ao.Idx, &(ao < j), ao * j, (*ao[j])[1], ao[j]->Retro, ao.InHouse(j));

   }
   if(ao.IfStars())
   {
     for(j = First_Star; j <= Last_Star; j++)
       if(ao == j)
         arq.Exec("INSERT INTO Obj (DataIdx, Id, Lon, Lat, Retro, InHouse) VALUES (%d, '%a', %f, %f, %b, %d)",
           ao.Idx, &(ao < j), ao * j, (*ao[j])[1], ao[j]->Retro, ao.InHouse(j));

   }
    if(afal)
    {
      AstroAspects aa(~ao);
      AspectFound* af;
      AstroFoundAspectsList& a = (AstroFoundAspectsList&)*afal;
      for(af = a.First(); af; af = a.Next())
      {
        char c;
        switch(af->AspectSide)
        {
          case Exact:      c = 'E'; break;
          case Applying:   c = 'A'; break;
          case Separating: c = 'S'; break;
        };
        as = ao < af->Oa;
        as += '-';
        as += ao < af->Ob;
        arq.Exec("INSERT INTO Asp (DataIdx, Objs, AspId, Orb, Pos) VALUES (%d, '%a', '%a', %f, '%c')",
         ao.Idx, &as, &(aa < af->Asp), af->DiffLong, c);
      }

    }
		arq.Commit();
    return ao.Idx;
}

void AstroChartList::UpdateData(AstroData* ad, bool placeonly)
{
  AstroChart* ac;
  for(ac = First(); ac; ac = Next())
  {
    if(ac->UpdateData(ad, placeonly))
			ac->Repaint();
  }
}

void AstroChartList::UpdateZif(Index idx, const char* zif)
{
  AstroChart* ac;
  for(ac = First(); ac; ac = Next())
  {
    if(ac->UpdateZif(idx, zif))
			ac->Repaint();
  }
}

void AstroChartList::UpdateRestrict(AstroRestrictions* ar)
{
  AstroChart* ac;
  for(ac = First(); ac; ac = Next())
  {
    if(ac->UpdateRestrict(ar))
			ac->Repaint();
  }
}

bool AstroChartList::CheckIdx(Index idx, bool country)
{
  AstroChart* ac;
  AstroObjs* ao;
  int i;
  for(ac = First(); ac; ac = Next())
  {
    for(i = 0; i < NBDATA; i++)
    {
      ao = (*ac)[i];
      if(ao && ((country ? ao->CountryIdx : ao->PlaceIdx) == idx))
				return TRUE;
    }
  }
  return FALSE;
}

void AstroChartList::ChangeIdx(Index oldidx, Index newidx)
{
  AstroChart* ac;
  AstroObjs* ao;
  int i;
  for(ac = First(); ac; ac = Next())
  {
    for(i = 0; i < NBDATA; i++)
    {
      ao = (*ac)[i];
      if(ao && ao->Idx == oldidx)
				ao->Idx = newidx;
    }
  }
}
