/***************************************************************************
                          astroprocess.cpp  -  description
                             -------------------
    begin                : Fri May 11 2001
    copyright            : (C) 2001 by Christophe Gros
    email                : skylendar@yahoo.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "astroprocess.h"
#include "astrostring.h"
#include "astroresources.h"
#include "astrointerface.h"
#include "astroutils.h"
#include "astrorestrictions.h"
#include "astrochartbase.h"
#include "aspscanlist.h"
#include "stopscan.h"

#ifdef KDE
#include "kapp.h"
#include <qprogressbar.h>
#endif
#include <cstdio>

extern const char* HousesCodes;

extern "C"
{
# include <sys/types.h>
# include <sys/wait.h>
# include <unistd.h>
# include <signal.h>
# include <stdlib.h>
# include "swephexp.h"
};

#define APP app.local8Bit()
#define ID  id.local8Bit()

const AstroString ScanClientStr(SCANCLIENT);

struct GetHousesStruct
{
  double cusps[13], asc, mc, fc, desc, vertex, east_point;
};

static int piddes[2], sigdes[2];
 
static void WritePid(pid_t pid)
{
  write(piddes[1], &pid, sizeof(pid_t));
}

static pid_t ReadPid()
{
  pid_t i;
  read(piddes[0], &i, sizeof(pid_t));
  return i;
}

static void WriteSig(char c)
{
  write(sigdes[1], &c, 1);
}

static char ReadSig()
{
  char c;
  read(sigdes[0], &c, 1);
  return c;
}

void Values::Set(int i, double d, bool r, double d1, double d2, double d3, double d4, double d5)
{
  if(i < 0 || i > Last_Star + 1) return;
  double *v = Data[i].V;
  v[0] = d;
  v[1] = d1;
  v[2] = d2;
  v[3] = d3;
  v[4] = d4;
  v[5] = d5;
  Data[i].Retro = r;
}

void Values::CSet(int i, double d, bool r)
{
  if(i < 0 || i > Last_Star + 1) return;
  double *v = Data[i].V;
  v[0] = d;
  Data[i].Retro = r;
}

AstroProcServer::AstroProcServer()
{
  int pgid;
  static char *argv[] = { ASTROPROCSESSERVER };
  setpgrp();
  pgid = getpgrp();
  pipe(piddes);
  pipe(sigdes);
  pid = fork();
  if(pid > 0)
  { // Proc server parent
    WritePid(getpid());
    while(1)
    {
      char c = ReadSig();
      int argc = 1;
      if(c == 'N')
      {
				pid_t p = fork();
				if(!p)
	  		{ // Comp server child
#ifdef KDE
	  			AstroString s = ASTROCOMPUTING;
	  			s += getpid();
	  			KApplication CmpServer(argc, argv, s.Data(), FALSE, FALSE);
	  			DCOPClient* dc = CmpServer.dcopClient();
	  			if(dc->attach())
	  			{
	    			dc->registerAs(s.Data(), FALSE);
	    			AstroComputing Ac;
	    			CmpServer.exec();
	  			}
#endif
				}
				else // Proc server parent
	  			WritePid(p);
      }
      else if(c == 'K')
				killpg(getpgrp(), 15);
    }
  }
  else // Proc server child
  {
    setpgid(0, pgid);
    pid = ReadPid();
  }
}

int AstroProcServer::NewCmpServer()
{
  pid_t p;
  WriteSig('N');
  p = ReadPid();
  return p;
}

void AstroProcServer::KillAll()
{
  WriteSig('K');
}

AstroInterface::AstroInterface(AstroString app, AstroString id) :
#ifdef KDE
    DCOPStub(APP, ID), AstroComputing_stub(APP, ID)
#endif
{
  ZoneInfoBuf[0] = 0;
}

AstroInterface* CreateInterface(bool cmpserver)
{
  int i;
  AstroString s = ASTROCOMPUTING;
  pid_t p = 1;
  if(cmpserver)
    p = AstroProcServer::NewCmpServer();
  if(p <= 0) return 0;
  s += p;
  AstroInterface* asi = new AstroInterface(cmpserver ? s : ASTROINTERFACE, COMPUTEOBJECT);
  if(!asi) return 0;
  for(i = 0; i <= 10000; i++)
  {
    asi->Sync();
    if(asi->ok())
      break;
  }
  if(i == 10000)
  {
    AstroMessage(i18n("Can't connect to computing server."));
    AstroProcServer::KillAll();
  }
  asi->KDESetSweDataPath(Asr->SWEphems);
  ARTransfer art(*Drs);
  asi->InitRestricts(art);
  ~art;
  return asi;
}

AstroComputing::AstroComputing(): QObject(), DCOPObject(COMPUTEOBJECT)
{
	int i;
	for(i = 0; i < NBDATA; i++)
		Vt[i] << V[i];
	Cscl.Init(this);
}

void AstroInterface::SetSweDataPath(const char* path)
{
  KDESetSweDataPath(path);
}

void AstroComputing::KDESetSweDataPath(QString path)
{
  swe_set_ephe_path((char *)path.data());
}

AstroComputing::~AstroComputing()
{
  int i;
  if(tzsave[0])
    setenv("TZ", tzsave, 1);
 else
    unsetenv("TZ");
  for(i = 0; i < NBDATA; i++)
    delete Ar[i];
}

void AstroInterface::SetTZ(const char *file, bool set)
{
  char buf[BUFSIZE];
  if(strcmp(file, ZoneInfoBuf))
  {
    strncpy(buf, Asr->ZPath, BUFSIZE);
    strncat(buf, file, BUFSIZE);
    KDESetTZ(buf, set);
    if(file)
      strncpy(ZoneInfoBuf, file, BUFSIZE);
  }
}

void AstroComputing::KDESetTZ(QString file, bool set)
{
  char *c, varbuf[80];
  if(file.data())
  {
    c = getenv("TZ");
    if(c)
      strncpy(tzsave, c, 80);
    varbuf[0] = ':';
    varbuf[1] = 0;
    strncat(varbuf, file.data(), 80);
    setenv("TZ", varbuf, 1);
    if(set)
      tzset();
  }
}

void AstroComputing::SetTZTime(double time, int day, int month, int year)
{
  struct tm tm;
  int h, m;
  h = (int)floor(time);
  m =  (int)(FRACT(fabs(time)) * 100.0 + ROUND / 600.0);
  tm.tm_sec = 0;
  tm.tm_min = m;
  tm.tm_hour = h;
  tm.tm_mday = day;
  tm.tm_mon = month - 1;
  tm.tm_year = year-1900;
  tt = mktime(&tm);
}

double AstroComputing::TimeOffset()
{
  struct tm *tmptrg;
  double dt, zheure, zminute;
  tmptrg =  gmtime(&tt);
  dt = difftime(mktime(tmptrg), tt);
  zminute = floor(dt/60.0);
  zheure = ceil(zminute / 60.0);
  zminute = floor(zminute - (zheure * 60.0));
  return -(zheure + zminute / 100.0);
}

bool AstroComputing::Isdst()
{
  struct tm *tmptrl;
  tmptrl = localtime(&tt);
  return tmptrl->tm_isdst ? TRUE : FALSE;
}

static void GetHouses(const struct DataComp& dc, char housecode, struct GetHousesStruct *ghs)
{
  double ascmc[10];
  (void)swe_houses_ex(dc.julday, dc.extraflags, dc.latitude, dc.longitude, (int)housecode, ghs->cusps, ascmc);
  ghs->asc        = ascmc[0];
  ghs->mc         = ascmc[1];
  ghs->fc         = Normalize(ascmc[1] + 180.0);
  ghs->desc       = Normalize(ascmc[0] + 180.0);
  ghs->vertex     = ascmc[3];
  ghs->east_point = ascmc[4];
}

static bool GetPlanet(int index, const struct DataComp& dc, const struct GetHousesStruct& ghs, double *array, char *serr, const _AstroRestrictions& arr, bool *retro = 0)
{
  double Su, Mo;
  int i = index;
  long l;

  if(index == North_Node || index == South_Node)
    index = (arr.TrueNode) ? SE_TRUE_NODE : SE_MEAN_NODE;
  if(index >= Fortune && index <= East_Point)
  {
    switch(index)
    {
      case Fortune: swe_calc_ut(dc.julday, Sun, dc.extraflags, array, serr);
				Su = *array;
				swe_calc_ut(dc.julday, Moon, dc.extraflags, array, serr);
				Mo = *array;
				*array = Normalize(ghs.asc - Su + Mo); break;
      case Vertex: *array = ghs.vertex; break;
      case East_Point: *array = ghs.east_point; break;
    }
    *(++array) = *(++array) = *(++array) = *(++array) = *(++array) = 0.0;
    if(retro)
      *retro = FALSE;
    return TRUE;
  }
  if(index > Black_Moon && index < Fortune)
    index++;
  l = swe_calc_ut(dc.julday, index, dc.extraflags, array, serr);
  if(l < 0)
    return FALSE;
  if(i == South_Node)
    *array = Normalize(*array + 180.0);
  if(retro)
  {
      double array2[6];
      if(i == South_Node)
      {
				*retro = TRUE;
				return TRUE;
		  }
      l = swe_calc_ut(dc.julday + ONEMINUTE, index, dc.extraflags, array2, serr);
      *retro = (*array > array2[0]) ? TRUE : FALSE;
  }
  return TRUE;
}

bool AstroComputing::Recompute(DCTransfert dct)
{
	struct DataComp dc ;
  dct > dc;
  const ComputeSubChart* csc = Cscl.Get(dc.cst);
  return csc ? csc->Compute(dc) : FALSE;
}

bool AstroComputing::Compute(const struct DataComp& dc)
{
  int i, flags = dc.extraflags;
  double array[6];
  bool retro;
  char buf[4];
  long l;
  struct GetHousesStruct ghs;
  const _AstroRestrictions& Arr = *Ar[dc.restr];
  Values& Vals = V[dc.ith];
  Vals.Julday = dc.julday;
  if(Arr.Sidereal > Tropical)
  {
    flags |= SEFLG_SIDEREAL;
    swe_set_sid_mode(Arr.Sidereal, 0, 0);
  }
  if(dc.gethouses)
  	GetHouses(dc,  HousesCodes[Arr.GetHouse()], &ghs);
  Vals.Set(Ascendant, ghs.asc);
  Vals.Set(MC, ghs.mc);
  Vals.Set(FC, ghs.fc);
  Vals.Set(Descendant, ghs.desc);
  for(i = House_1; i <= House_12; i++)
    Vals.Set(i, ghs.cusps[i - House_1 +1]);
  if(Arr == Vertex)
    Vals.Set(Vertex, ghs.vertex);
  if(Arr == East_Point)
    Vals.Set(East_Point, ghs.east_point);
  for(i = Sun; i <= Fortune; i++)
  {
    if(Arr == i)
    {
      if(GetPlanet(i, dc, ghs, array, serr, Arr, &retro))
				Vals.Set(i, array[0], retro, array[1], array[2], array[3], array[4], array[5]);
			else
				return FALSE;
    }
  }
  if(Arr.IfExotic())
  {
    for(i = First_Exotic; i <= Last_Exotic; i++)
      if(Arr == i)
      {
				if(GetPlanet(i, dc, ghs, array, serr, Arr, &retro))
					Vals.Set(i, array[0], retro, array[1], array[2], array[3], array[4], array[5]);
				 else
	  			return FALSE;
      }
  }
  if(Arr.IfStars())
  {
    for(i = First_Star; i <= Last_Star; i++)
      if(Arr == i)
      {
				sprintf(buf, "%d", i - First_Star + 1);
				l = swe_fixstar_ut(buf, dc.julday, dc.extraflags, array, serr);
				if(l < 0)
					return FALSE;
				Vals.Set(i, array[0], FALSE, array[1], array[2], array[3], array[4], array[5]);
      }
  }
  return TRUE;
}

VTransfer AstroComputing::GetValues(int slot) const
{
	return Vt[slot];
}

void AstroComputing::SetRestrict(ARTransfer art, int i)
{
  art > *Ar[i];
}

QString AstroComputing::KDEGetError()
{
  return serr;
}

void AstroInterface::SetRestrict(const AstroRestrictions* ar, int i)
{
  if(i < 0 || i >= NBDATA) return;
  Art = *SC(_AstroRestrictions*, (AstroRestrictions*)ar);
  AstroComputing_stub::SetRestrict(Art, i);
}

void AstroInterface::GetError(AstroString& as)
{
  as = KDEGetError();
}

void AstroComputing::InitRestricts(ARTransfer art)
{
  int i;
  _AstroRestrictions ar;
  art > ar;
  for(i = 0; i < NBDATA; i++)
  {
    Ar[i] = new _AstroRestrictions;
    *Ar[i] = ar;
  }
}

void AstroComputing::Kill()
{
  exit(0);
}

void AstroComputing::Sync()
{
  // Does nothing !
}

void AstroComputing::ScanAsp(DCTransfert dct, double origjd, QString scserver, int scanflags)
{
	struct DataComp dc;
	double j1, j2, last;
	Values v1;
	AspScanParams Asp;
	Asp.Get(scanflags);
	AstroString  as(ASTROINTERFACE), sc(scserver);
	AstroFoundAspectsTable afat1, afat2, *afp1, *afp2, *afp3;
	ScanInterface sci(as, sc);
	dct > dc;
	IfScan = TRUE;
	dc.extraflags |= SEFLG_SPEED;
	const Values& orig = V[dc.first];
	const _AstroAspects aa(Ar[dc.ith]);
	const _AstroRestrictions &ar1 = *Ar[dc.first], &ar2 = *Ar[dc.ith];
	const int mins = ((24 * 60) / 4);
	last = dc.julday;
	afp1 = &afat1;
	afp2 = &afat2;
	j1 = origjd;
	v1 = V[dc.ith];
	j2 = IncTime(j1, mins, 0, 0, 0, 0);
	dc.julday = j2;
	if(!Compute(dc)) return;
	afp1->GetTable(aa, orig, V[dc.ith], TRUE, ar1, ar2);
	v1 = V[dc.ith];
	j1 = j2;
	while(j2 < last)
	{
		sci.Progress((int)((j2 - origjd)  * 100.0 / (last - origjd)));
		j2 = IncTime(j1, mins, 0, 0, 0, 0);
		if(j2 > last) j2 = last;
		dc.julday = j2;
		if(!Compute(dc)) return;
		afp2->GetTable(aa, orig, V[dc.ith], TRUE, ar1, ar2);
		TestAsp(aa, *afp1, *afp2, v1, V[dc.ith], j1, dc, sci);
		if(!IfScan) break;
		v1 = V[dc.ith];
		j1 = j2;
		afp3 = afp1; afp1 = afp2; afp2 = afp3;
	}
	sci.EndScan();
}

void AstroComputing::TestAsp(const _AstroAspects& aa, const AstroFoundAspectsTable& afat1,
	const AstroFoundAspectsTable& afat2, const Values& v1, const Values& v2,
	double j1, const struct DataComp& dc, ScanInterface& sci)
{
	int i, j, rd, max_x, max_y;
	double r, v, x, p1, p2;
	const double mins = ((24 * 60) / 4) * ONEMINUTE;
	enum Aspects a1, a2;
	bool change;
	const Values &orig = V[dc.first];
	rd = -1;
	afat2.GetBounds(&max_x, &max_y);
	for(i = Sun; i <= max_x; i++)
	{
#ifdef KDE
		//if(qApp->hasPendingEvents())
		qApp->processEvents(10);
		if(!IfScan) return;
#endif
		for(j = Sun; j <= max_y; j++)
		{
			a1 = afat1.Get(i, j);
			a2 = afat2.Get(i, j);
			if(a1 != Unknown_Aspect && a2 != Unknown_Aspect && a1 == a2)
			{
				p1 = v1.Get(j);
				p2 = v2.Get(j);
				change = (v1.Retro(j) + v2.Retro(j)) == 1 ? TRUE : FALSE;
				if(p2 < p1 && !v2.Retro(j) && !change) p2 += 360.0;
				if(p2 > p1 && v2.Retro(j) && !change)  p1 += 360.0;
				if(change && rd != j)
				{
					double s1 = v1.Get(j, 3), acc = (v2.Get(j, 3) - s1) / .246, z = -s1 / acc;
					sci.GetRetro(j, s1 >= 0.0 ? TRUE : FALSE, j1 + z);
					rd = j;
				}
				r = Normalize(orig.Get(i) - (aa * a1));
				if(p2 >= r && p1 <= r)
				{
					v = (p2 - p1) / mins;
					x = j1 + ((r - p1) / v);
					//printf(" - i = %d j = %d p1 = %f p2 = %f r = %f orig = %f\n", i, j, p1, p2, r, orig.Get(i));
					sci.GetAspect(j, p1 * (v1.Retro(j) ? -1.0 : 1.0), i, orig.Get(i) * (orig.Retro(i) ? -1.0 : 1.0), a1, x);
				}
				else
				{
					r = orig.Get(i) + (aa * a1);
					if(p2 >= r && p1 <= r)
					{
						v = (p2 - p1) / mins;
						x = j1 + (r - p1) / v;
						//printf(" + i = %d j = %d p1 = %f p2 = %f r = %f\n", i, j, p1, p2, r);
						sci.GetAspect(j, p1 * (v1.Retro(j) ? -1.0 : 1.0), i, orig.Get(i) * (orig.Retro(i) ? -1.0 : 1.0), a1, x);
					}
				}
			}
		}
 }
}

void AstroComputing::StopScan()
{
	IfScan = FALSE;
}

ScanClient::ScanClient(AspScanList* ascl, int id): QObject(), DCOPObject((ScanClientStr + id).Data())
{
	Ascl = ascl;
	End = FALSE;
}

ScanClient::~ScanClient()
{
}

void ScanClient::GetAspect(int i, double ipos, int j, double jpos, int asp, double date)
{
	Ascl->DrawDate(i, ipos, j, jpos, asp, date);
}

void ScanClient::EndScan()
{
	End = TRUE;
}

void ScanClient::GetRetro(int i, double ipos, double date)
{
	Ascl->DrawChange(i, ipos, date);
}

void ScanClient::Progress(int i)
{
	emit Ascl->Ss->ProgressBar->setProgress(i);
}

ScanInterface::ScanInterface(AstroString app, AstroString id)
#ifdef KDE
    : DCOPStub(APP, ID), ScanClient_stub(APP, ID)
#endif
{
}

bool ComputeSubChart::Compute(const struct DataComp& dc) const
{
	return Ac->Compute(dc);
}

Values& ComputeSubChart::V(int i) const
{
	return Ac->V[i];
}

const _AstroRestrictions& ComputeSubChart::Ar(int i) const
{
	return *Ac->Ar[i];
}

class ComputeProgression: public ComputeSubChart
{
public:

	ComputeProgression(AstroComputing* ac): ComputeSubChart(ac) {}

  bool Compute(const struct DataComp& dc) const
  {
    int  y, mo, d, h, m;
    double jd = V(dc.first).Julday;
    struct DataComp dc1 = dc;
    Diffs(jd, dc.julday, d, h, m);
    y = (int)(d / 365.25);
    d -= (int)(y * 365.25);
    mo = (int)(d / 30.4375);
    d -= (int)(mo * 30.4375);
    dc1.julday = Progress(jd, y, mo, d);
    V(dc.ith).Julday = dc.julday;
    return ComputeSubChart::Compute(dc1);
  }
};

class ComputeDirection: public ComputeSubChart
{
public:

	ComputeDirection(AstroComputing* ac): ComputeSubChart(ac) {}

  bool Compute(const struct DataComp& dc) const
  {
    int i, y, mo, d, h, m, house[Undefined_Obj];
    const _AstroRestrictions& ar = Ar(dc.first);
    struct DataComp dc1 = dc;
    if(dc.ith == dc.first) return FALSE;
		if(!ar.IfHouses() || !Ar(dc.ith).IfHouses()) return FALSE;
    const Values& vf = V(dc.first);
    Values& vi = V(dc.ith);
    double jd1 = vf.Julday, pos[Undefined_Obj];
    for(i = Sun; i <= Last_Planet; i++)
      if(ar == i)
				house[i] = vf.InHouse(i, &pos[i]);
    for(i = Ascendant; i <= House_12; i++)
      if(ar == i)
				house[i] = vf.InHouse(i, &pos[i]);
    if(ar.IfExotic())
      for(i = First_Exotic; i <= Last_Exotic; i++)
				if(ar == i)
					house[i] = vf.InHouse(i, &pos[i]);
    if(ar.IfStars())
      for(i = First_Star; i <= Last_Star; i++)
				if(ar == i)
					house[i] = vf.InHouse(i, &pos[i]);
    Diffs(dc.julday, jd1, d, h, m);
    y  = (int)(d / 365.25);
    d -= (int)(y * 365.25);
    mo = (int)(d / 30.4375);
    d -= (int)(mo * 30.4375);
    dc1.julday = Direct(jd1, y, mo);
    Diffs(dc1.julday, jd1, d, h, m);
    vi.Julday = dc.julday;
    if(!ComputeSubChart::Compute(dc1)) return FALSE;
    for(i = Sun; i <= Last_Planet; i++)
      if(ar == i)
				vi.PutInHouse(i, house[i], pos[i]);
    for(i = Ascendant; i <= House_12; i++)
      if(ar == i)
				vi.PutInHouse(i, house[i], pos[i]);
    if(ar.IfExotic())
      for(i = First_Exotic; i <= Last_Exotic; i++)
				if(ar == i)
					vi.PutInHouse(i, house[i], pos[i]);
    if(ar.IfStars())
      for(i = First_Star; i <= Last_Star; i++)
				if(ar == i)
					vi.PutInHouse(i, house[i], pos[i]);
  return TRUE;
  }
};

ComputeSubChartList::ComputeSubChartList()
{
	int i;
	for(i = NoSt; i <= Last_SubType; i++)
		Cscl[i] = 0;
}

void ComputeSubChartList::Init(AstroComputing* ac)
{
	Cscl[Single]      = new ComputeSubChart(ac);
	Cscl[Transit]     = new ComputeSubChart(ac);
	Cscl[Progression] = new ComputeProgression(ac);
	Cscl[Direction]   = new ComputeDirection(ac);
}

ComputeSubChartList::~ComputeSubChartList()
{
	int i;
	for(i = NoSt; i <= Last_SubType; i++)
		if(Cscl[i])
			delete Cscl[i];
}

int Values::InHouse(int i, double *f) const
{
  int j;
  double x = Data[i][0], x1, x2, x3;
  for(j = 0; j < 12; j++)
  {
    x1 = Data[j + House_1][0];
    x2 = Data[j + House_1 + 1][0];
    if(x2 < x1)
    {
      x2 += 360.0;
      x3 = x + 360.0;
      if(x3 >= x1 && x3 <= x2)
				x = x3;
    }
    if(x >= x1 && x <= x2)
    {
      if(f)
				*f = 1000.0 * (x - x1) / (x2 - x1);
      return j + 1;
    }
  }
  return 0;
}

void Values::PutInHouse(int index, int house, double f)
{
  double x1, x2;
  x1 = Data[house + House_1 - 1][0];
  x2 = Data[house + House_1][0];
  if(x2 < x1)
  {
    x2 += 360.0;
  }
  f = Normalize(x1 + (f * (x2 - x1) / 1000.0));
  Data[index].V[0] = f;
  //Vals.CSet(index, f, Vals.Retro(index));
}
