BEGIN;

CREATE TABLE VData -- table containing additional refernce for values storing
(
	DataIdx Int4 NOT NULL, -- Data row index
	RefIdx  Int4 NOT NULL, -- Original data row index
	RestIdx Int4 NOT NULL, -- Restrictions set index
	Type CHAR NOT NULL -- 'S' = single 'T' = transit 'D' = directions 'P' = progressions
);

CREATE INDEX VData_Idx_Key ON VData(DataIdx); 

CREATE VIEW VDV AS SELECT * FROM Data, VData WHERE Data.Idx = VData.DataIdx; -- This view displays all the data with saved values

CREATE TABLE Obj -- Objects values
(
	DataIdx Int4 NOT NULL, -- Data row index
	Id CHAR(7) NOT NULL, -- Obj id, e.g. 'Sat' for Saturn
	Lon NUMERIC(8, 3) NOT NULL, -- Longitude in decimal
	Lat NUMERIC(8, 3) NOT NULL, -- latitude in decimal
	Retro BOOL NOT NULL, -- If retro
	InHouse Int2 NOT NULL -- House number
);

CREATE INDEX Obj_DataIdx_Key ON Obj(DataIdx);

CREATE TABLE Asp -- Aspects values
(
	DataIdx Int4 NOT NULL, -- Data row index
	Objs CHAR(15) NOT NULL, -- Objs in aspect .e.g. 'Mar-Ven' for Mars-Venus aspect  
	AspId CHAR(3) NOT NULL, -- Aspect id .e.g. 'Opp' for opposition
	Orb NUMERIC(5, 3) NOT NULL, -- Local orb
	Pos CHAR NOT NULL -- If exact 'E', applying 'A', or separating 'S', aspect
);

CREATE INDEX Asp_DataIdx_Key ON Asp(DataIdx);

COMMIT;

CREATE OR REPLACE FUNCTION Del_VData () RETURNS OPAQUE AS '
    BEGIN
	DELETE FROM Obj WHERE DataIdx = OLD.DataIdx;
	DELETE FROM Asp WHERE DataIdx = OLD.DataIdx;
        RETURN OLD;
    END;
' LANGUAGE 'plpgsql';

CREATE OR REPLACE FUNCTION Del_Data () RETURNS OPAQUE AS '
    DECLARE
	r RECORD;
    BEGIN
	SELECT INTO r RefIdx FROM VDV WHERE RefIdx = OLD.Idx AND DataIdx <> RefIdx;
	IF FOUND THEN
		RAISE EXCEPTION ''Cannot delete this data set. It is referenced by other data sets''; 
		RETURN NULL;
	END IF;
	RAISE NOTICE ''DELETE Data Idx = %'', OLD.Idx;
	DELETE FROM VData WHERE DataIdx = OLD.Idx;
        RETURN OLD;
    END;
' LANGUAGE 'plpgsql';

CREATE OR REPLACE FUNCTION Del_Restrictions() RETURNS OPAQUE AS '
    BEGIN
	RAISE NOTICE ''DELETE Restrictions Idx = %'', OLD.Idx;
	UPDATE VData SET RestIdx = 0 WHERE RestIdx = OLD.Idx;
	RETURN OLD;
    END;
' LANGUAGE 'plpgsql';

CREATE OR REPLACE FUNCTION IfSign(Real, Integer) RETURNS Boolean AS '
    DECLARE
	angle ALIAS FOR $1;
	sn ALIAS FOR $2;
	d INTEGER;
    BEGIN
	d := FLOOR(angle / 30.0) + 1;
	IF d = sn THEN
		RETURN TRUE;
	ELSE
		RETURN FALSE;
	END IF;
    END;
' LANGUAGE 'plpgsql';

CREATE OR REPLACE FUNCTION TObj(Int4, VARCHAR, Integer) RETURNS Boolean AS '
   DECLARE
	r RECORD;
	o ALIAS FOR $2; 
   BEGIN
	SELECT INTO r DataIdx, Lon FROM Obj WHERE DataIdx = $1 AND Id ~* o AND IfSign(Lon, $3);
	IF FOUND THEN
		RETURN TRUE;
	ELSE
		RETURN FALSE;
	END IF;
    END;
' LANGUAGE 'plpgsql';
 
CREATE OR REPLACE FUNCTION IfAsp(VARCHAR, VARCHAR, VARCHAR) RETURNS Boolean AS '
    DECLARE
	asp VARCHAR;
	a VARCHAR;
	b VARCHAR;
	per CONSTANT VARCHAR := ''%'';
	da CONSTANT VARCHAR := ''-'';
    BEGIN
	asp := rtrim($1, '' '');
	a := rtrim($2, '' '');
	b := rtrim($3, '' '');
	IF a = '''' THEN
		IF asp LIKE per || b || per THEN
			RETURN TRUE;
		END IF;
	ELSIF b = '''' THEN
		IF asp LIKE per || a || per  THEN
			RETURN TRUE;
		END IF;
	ELSIF asp LIKE (a || da || b) OR asp LIKE (b || da || a) THEN
			RETURN TRUE;
	END IF;
    	RETURN FALSE;
    END;
' LANGUAGE 'plpgsql';

CREATE OR REPLACE FUNCTION TAsp(Int4, VARCHAR, VARCHAR, VARCHAR) RETURNS Boolean AS '
   DECLARE
	r RECORD;
	ai ALIAS FOR $2;
	o1 ALIAS FOR $3;
	o2 ALIAS FOR $4; 
   BEGIN
	SELECT INTO r DataIdx, AspId, Objs FROM Asp WHERE DataIdx = $1 AND AspId ~* ai AND IfAsp(Objs, o1, o2);
	IF FOUND THEN
		RETURN TRUE;
	ELSE
		RETURN FALSE;
	END IF;
    END;
' LANGUAGE 'plpgsql';

CREATE OR REPLACE FUNCTION TRetr(Int4, VARCHAR) RETURNS Boolean AS '
   DECLARE
	ai ALIAS FOR $2;
	r RECORD;
   BEGIN
	SELECT INTO r Id, Retro FROM Obj WHERE DataIdx = $1 AND Id ~* ai;
	IF FOUND THEN 
		RETURN r.Retro;
	END IF;
	RETURN FALSE;
   END;
' LANGUAGE 'plpgsql';

CREATE OR REPLACE FUNCTION THouse(Int4, VARCHAR, Integer) RETURNS Boolean AS '
   DECLARE
	ai ALIAS FOR $2;
	r RECORD;
   BEGIN
	SELECT INTO r InHouse FROM Obj WHERE DataIdx = $1 AND Id ~* ai;
	IF FOUND THEN
		IF r.InHouse = $3 THEN
			RETURN TRUE;
		END IF;
	END IF;
	RETURN FALSE;
   END;
' LANGUAGE 'plpgsql';

BEGIN;

CREATE TRIGGER Del_VData_Trigger BEFORE DELETE ON VData
    FOR EACH ROW EXECUTE PROCEDURE Del_VData();

CREATE TRIGGER Del_Data_Trigger BEFORE DELETE ON Data
    FOR EACH ROW EXECUTE PROCEDURE Del_Data();

CREATE TRIGGER Del_Restrictions_Trigger BEFORE DELETE ON Restrictions
    FOR EACH ROW EXECUTE PROCEDURE Del_Restrictions();

COMMIT;